# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, List, NamedTuple, Optional
from urllib import request
from zipfile import ZipFile

from pandas.tseries.frequencies import to_offset

from gluonts.dataset import DatasetWriter
from gluonts.dataset.common import MetaData, TrainDatasets
from gluonts.dataset.field_names import FieldName
from gluonts.gluonts_tqdm import tqdm

from ._tsf_reader import TSFReader, frequency_converter
from ._util import metadata, request_retrieve_hook


class Dataset(NamedTuple):
    file_name: str
    record: str
    ROOT: str = "https://zenodo.org/record"

    @property
    def url(self):
        return f"{self.ROOT}/{self.record}/files/{self.file_name}"

    def download(self, path: Path):
        file_path = path / self.file_name
        with tqdm(
            [],
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            miniters=5,
            desc=f"Download {self.file_name}:",
        ) as _tqdm:
            request.urlretrieve(
                self.url,
                filename=file_path,
                reporthook=request_retrieve_hook(_tqdm),
            )
        return file_path


datasets = {
    "kaggle_web_traffic_with_missing": Dataset(
        file_name="kaggle_web_traffic_dataset_with_missing_values.zip",
        record="4656080",
    ),
    "kaggle_web_traffic_without_missing": Dataset(
        file_name="kaggle_web_traffic_dataset_without_missing_values.zip",
        record="4656075",
    ),
    "kaggle_web_traffic_weekly": Dataset(
        file_name="kaggle_web_traffic_weekly_dataset.zip",
        record="4656664",
    ),
    "m1_yearly": Dataset(file_name="m1_yearly_dataset.zip", record="4656193"),
    "m1_quarterly": Dataset(
        file_name="m1_quarterly_dataset.zip", record="4656154"
    ),
    "m1_monthly": Dataset(
        file_name="m1_monthly_dataset.zip", record="4656159"
    ),
    "nn5_daily_with_missing": Dataset(
        file_name="nn5_daily_dataset_with_missing_values.zip",
        record="4656110",
    ),
    "nn5_daily_without_missing": Dataset(
        file_name="nn5_daily_dataset_without_missing_values.zip",
        record="4656117",
    ),
    "nn5_weekly": Dataset(
        file_name="nn5_weekly_dataset.zip", record="4656125"
    ),
    "tourism_monthly": Dataset(
        file_name="tourism_monthly_dataset.zip",
        record="4656096",
    ),
    "tourism_quarterly": Dataset(
        file_name="tourism_quarterly_dataset.zip",
        record="4656093",
    ),
    "tourism_yearly": Dataset(
        file_name="tourism_yearly_dataset.zip",
        record="4656103",
    ),
    "cif_2016": Dataset(
        file_name="cif_2016_dataset.zip",
        record="4656042",
    ),
    "london_smart_meters_without_missing": Dataset(
        file_name="london_smart_meters_dataset_without_missing_values.zip",
        record="4656091",
    ),
    "wind_farms_without_missing": Dataset(
        file_name="wind_farms_minutely_dataset_without_missing_values.zip",
        record="4654858",
    ),
    "car_parts_without_missing": Dataset(
        file_name="car_parts_dataset_without_missing_values.zip",
        record="4656021",
    ),
    "dominick": Dataset(
        file_name="dominick_dataset.zip",
        record="4654802",
    ),
    "fred_md": Dataset(
        file_name="fred_md_dataset.zip",
        record="4654833",
    ),
    "pedestrian_counts": Dataset(
        file_name="pedestrian_counts_dataset.zip",
        record="4656626",
    ),
    "hospital": Dataset(
        file_name="hospital_dataset.zip",
        record="4656014",
    ),
    "covid_deaths": Dataset(
        file_name="covid_deaths_dataset.zip",
        record="4656009",
    ),
    "kdd_cup_2018_without_missing": Dataset(
        file_name="kdd_cup_2018_dataset_without_missing_values.zip",
        record="4656756",
    ),
    "weather": Dataset(
        file_name="weather_dataset.zip",
        record="4654822",
    ),
    "australian_electricity_demand": Dataset(
        file_name="australian_electricity_demand_dataset.zip",
        record="4659727",
    ),
    "electricity_hourly": Dataset(
        file_name="electricity_hourly_dataset.zip",
        record="4656140",
    ),
    "electricity_weekly": Dataset(
        file_name="electricity_weekly_dataset.zip",
        record="4656141",
    ),
    "rideshare_without_missing": Dataset(
        file_name="rideshare_dataset_without_missing_values.zip",
        record="5122232",
    ),
    "saugeenday": Dataset(
        file_name="saugeenday_dataset.zip",
        record="4656058",
    ),
    "solar_10_minutes": Dataset(
        file_name="solar_10_minutes_dataset.zip",
        record="4656144",
    ),
    "solar_weekly": Dataset(
        file_name="solar_weekly_dataset.zip",
        record="4656151",
    ),
    "sunspot_without_missing": Dataset(
        file_name="sunspot_dataset_without_missing_values.zip",
        record="4654722",
    ),
    "temperature_rain_without_missing": Dataset(
        file_name="temperature_rain_dataset_without_missing_values.zip",
        record="5129091",
    ),
    "vehicle_trips_without_missing": Dataset(
        file_name="vehicle_trips_dataset_without_missing_values.zip",
        record="5122537",
    ),
}


def convert_data(
    data: List[Dict],
    train_offset: int,
    default_start_timestamp: Optional[str] = None,
):
    train_data = []
    test_data = []
    for i, data_entry in tqdm(
        enumerate(data), total=len(data), desc="creating json files"
    ):
        # Convert the data to a GluonTS dataset...
        # - `default_start_timestamp` is required for some datasets which
        #   are not listed here since some datasets do not define start
        #   timestamps
        # - `item_id` is added for all datasets ... many datasets provide
        #   the "series_name"
        test_data.append(
            {
                "target": data_entry["target"],
                "start": str(
                    data_entry.get("start_timestamp", default_start_timestamp)
                ),
                "item_id": data_entry.get("series_name", i),
                "feat_static_cat": [i],
            }
        )

        train_data.append(
            {
                "target": data_entry["target"][:-train_offset],
                "start": str(
                    data_entry.get("start_timestamp", default_start_timestamp)
                ),
                "item_id": data_entry.get("series_name", i),
                "feat_static_cat": [i],
            }
        )

    return train_data, test_data


def generate_forecasting_dataset(
    dataset_path: Path,
    dataset_name: str,
    dataset_writer: DatasetWriter,
    prediction_length: Optional[int] = None,
):
    dataset = datasets[dataset_name]
    dataset_path.mkdir(exist_ok=True)

    with TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)

        with ZipFile(dataset.download(temp_path)) as archive:
            archive.extractall(path=temp_path)

        # only one file is exptected
        reader = TSFReader(temp_path / archive.namelist()[0])
        meta, data = reader.read()

    freq = frequency_converter(meta.frequency)
    if prediction_length is None:
        if hasattr(meta, "forecast_horizon"):
            prediction_length = int(meta.forecast_horizon)
        else:
            prediction_length = default_prediction_length_from_frequency(freq)

    # Impute missing start dates with unix epoch and remove time series whose
    # length is less than or equal to the prediction length
    data = [
        {**d, "start_timestamp": d.get("start_timestamp", "1970-01-01")}
        for d in data
        if len(d[FieldName.TARGET]) > prediction_length
    ]
    train_data, test_data = convert_data(data, prediction_length)

    meta = MetaData(
        **metadata(
            cardinality=len(data),
            freq=freq,
            prediction_length=prediction_length,
        )
    )

    dataset = TrainDatasets(metadata=meta, train=train_data, test=test_data)
    dataset.save(
        path_str=str(dataset_path), writer=dataset_writer, overwrite=True
    )


def default_prediction_length_from_frequency(freq: str) -> int:
    prediction_length_map = {
        "T": 60,
        "H": 48,
        "D": 30,
        "W-SUN": 8,
        "M": 12,
        "Y": 4,
    }
    try:
        freq = to_offset(freq).name
        return prediction_length_map[freq]
    except KeyError as err:
        raise ValueError(
            f"Cannot obtain default prediction length from frequency `{freq}`."
        ) from err



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import copy
import tarfile
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

T = TypeVar("T")


def copy_with(obj: T, **kwargs) -> T:
    """Return copy of `obj` and update attributes of the copy using `kwargs`.

    ::

        @dataclass
        class MyClass:
            value: int

        a = MyClass(1)
        b = copy_with(a, value=2)

        assert a.value == 1
        assert b.value == 2

    """

    new_obj = copy.copy(obj)

    for name, value in kwargs.items():
        setattr(new_obj, name, value)

    return new_obj


if TYPE_CHECKING:
    lazy_property = property

else:

    def lazy_property(method):
        """Property that is lazily evaluated.

        This is the same as::

            @property
            @lru_cache(1)
            def my_property(self):
                ...

        In addition to be more concise, `lazy_property` also works with mypy,
        since it poses to be just `property` when type checked.

        This implementation follows the recipe from the `functools`
        documentation, and mimics `functools.cached_property` which was
        introduced in Python 3.8.
        """
        return property(lru_cache(1)(method))


def will_extractall_into(tar: tarfile.TarFile, path: Path) -> None:
    """
    Check that the content of ``tar`` will be extracted within ``path``
    upon calling ``extractall``.

    Raise a ``PermissionError`` if not.
    """
    path = Path(path).resolve()

    for member in tar.getmembers():
        member_path = (path / member.name).resolve()

        try:
            member_path.relative_to(path)
        except ValueError:
            raise PermissionError(f"'{member.name}' extracts out of target.")


def safe_extractall(
    tar: tarfile.TarFile,
    path: Path = Path("."),
    members=None,
    *,
    numeric_owner=False,
):
    """
    Safe wrapper around ``TarFile.extractall`` that checks all destination
    files to be strictly within the given ``path``.
    """
    will_extractall_into(tar, path)
    tar.extractall(path, members, numeric_owner=numeric_owner)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Optional, Tuple

import mxnet as mx
import numpy as np

from gluonts.core.component import validated
from gluonts.mx import Tensor
from gluonts.mx.linalg_util import jitter_cholesky
from gluonts.mx.util import _broadcast_param, make_nd_diag

from . import Distribution, Gaussian, MultivariateGaussian
from .distribution import getF


class ParameterBounds:
    @validated()
    def __init__(self, lower, upper) -> None:
        assert (
            lower <= upper
        ), "lower bound should be smaller or equal to upper bound"
        self.lower = lower
        self.upper = upper


def _safe_split(x, num_outputs, axis, squeeze_axis, *args, **kwargs):
    """
    A type-stable wrapper around mx.nd.split.

    Currently mx.nd.split behaves weirdly if num_outputs=1:

        a = mx.nd.ones(shape=(1, 1, 2))
        l = a.split(axis=1, num_outputs=1, squeeze_axis=False)
        type(l)  # mx.NDArray
        l.shape  # (1, 1, 2)

    Compare that with the case num_outputs=2:

        a = mx.nd.ones(shape=(1, 2, 2))
        l = a.split(axis=1, num_outputs=2, squeeze_axis=False)
        type(l)  # list
        len(l)  # 2
        l[0].shape  # (1, 1, 2)

    This wrapper makes the behavior consistent by always returning a list
    of length num_outputs, whose elements will have one less axis than x
    in case x.shape[axis]==num_outputs and squeeze_axis==True, and the same
    number of axes as x otherwise.
    """
    if num_outputs > 1:
        return x.split(
            axis=axis,
            num_outputs=num_outputs,
            squeeze_axis=squeeze_axis,
            *args,
            **kwargs
        )
    return [x.squeeze(axis=axis)] if squeeze_axis else [x]


class LDS(Distribution):
    r"""
    Implements Linear Dynamical System (LDS) as a distribution.

    The LDS is given by

    .. math::
        z_t = A_t l_{t-1} + b_t + \epsilon_t \\
        l_t = C_t l_{t-1} + g_t \nu

    where

    .. math::
        \epsilon_t = N(0, S_v) \\
        \nu = N(0, 1)

    :math:`A_t`, :math:`C_t` and :math:`g_t` are the emission, transition and
    innovation coefficients respectively. The residual terms are denoted
    by :math:`b_t`.

    The target :math:`z_t` can be :math:`d`-dimensional in which case

    .. math::
        A_t \in R^{d \times h}, b_t \in R^{d}, C_t \in R^{h \times h}, g_t \in R^{h}

    where :math:`h` is dimension of the latent state.

    Parameters
    ----------
    emission_coeff
        Tensor of shape (batch_size, seq_length, obs_dim, latent_dim)
    transition_coeff
        Tensor of shape (batch_size, seq_length, latent_dim, latent_dim)
    innovation_coeff
        Tensor of shape (batch_size, seq_length, latent_dim)
    noise_std
        Tensor of shape (batch_size, seq_length, obs_dim)
    residuals
        Tensor of shape (batch_size, seq_length, obs_dim)
    prior_mean
        Tensor of shape (batch_size, latent_dim)
    prior_cov
        Tensor of shape (batch_size, latent_dim, latent_dim)
    latent_dim
        Dimension of the latent state
    output_dim
        Dimension of the output
    seq_length
        Sequence length
    F
    """  # noqa: E501

    @validated()
    def __init__(
        self,
        emission_coeff: Tensor,
        transition_coeff: Tensor,
        innovation_coeff: Tensor,
        noise_std: Tensor,
        residuals: Tensor,
        prior_mean: Tensor,
        prior_cov: Tensor,
        latent_dim: int,
        output_dim: int,
        seq_length: int,
    ) -> None:
        self.latent_dim = latent_dim
        self.output_dim = output_dim
        self.seq_length = seq_length

        # Split coefficients along time axis for easy access
        # emission_coef[t]: (batch_size, obs_dim, latent_dim)
        self.emission_coeff = _safe_split(
            emission_coeff,
            axis=1,
            num_outputs=self.seq_length,
            squeeze_axis=True,
        )

        # innovation_coef[t]: (batch_size, latent_dim)
        self.innovation_coeff = _safe_split(
            innovation_coeff,
            axis=1,
            num_outputs=self.seq_length,
            squeeze_axis=False,
        )

        # transition_coeff: (batch_size, latent_dim, latent_dim)
        self.transition_coeff = _safe_split(
            transition_coeff,
            axis=1,
            num_outputs=self.seq_length,
            squeeze_axis=True,
        )

        # noise_std[t]: (batch_size, obs_dim)
        self.noise_std = _safe_split(
            noise_std, axis=1, num_outputs=self.seq_length, squeeze_axis=True
        )

        # residuals[t]: (batch_size, obs_dim)
        self.residuals = _safe_split(
            residuals, axis=1, num_outputs=self.seq_length, squeeze_axis=True
        )

        self.prior_mean = prior_mean
        self.prior_cov = prior_cov

    @property
    def F(self):
        return getF(self.prior_mean)

    @property
    def batch_shape(self) -> Tuple:
        return self.emission_coeff[0].shape[:1] + (self.seq_length,)

    @property
    def event_shape(self) -> Tuple:
        return (self.output_dim,)

    @property
    def event_dim(self) -> int:
        return 2

    def log_prob(
        self,
        x: Tensor,
        scale: Optional[Tensor] = None,
        observed: Optional[Tensor] = None,
    ):
        """
        Compute the log probability of observations.

        This method also returns the final state of the system.

        Parameters
        ----------
        x
            Observations, shape (batch_size, seq_length, output_dim)
        scale
            Scale of each sequence in x, shape (batch_size, output_dim)
        observed
            Flag tensor indicating which observations are genuine (1.0) and
            which are missing (0.0)

        Returns
        -------
        Tensor
            Log probabilities, shape (batch_size, seq_length)
        Tensor
            Final mean, shape (batch_size, latent_dim)
        Tensor
            Final covariance, shape (batch_size, latent_dim, latent_dim)
        """
        if scale is not None:
            x = self.F.broadcast_div(x, scale.expand_dims(axis=1))
        # TODO: Based on form of the prior decide to do either filtering
        #   or residual-sum-of-squares
        log_p, final_mean, final_cov = self.kalman_filter(x, observed)
        if scale is not None:
            F = self.F
            # log_abs_det_jac: sum over all output dimensions.
            ladj = -F.sum(F.log(F.abs(scale)), axis=-1, keepdims=True)

            # Sum `ladj` over all time steps.
            log_p = F.broadcast_add(log_p, ladj)
        return log_p, final_mean, final_cov

    def kalman_filter(
        self, targets: Tensor, observed: Tensor
    ) -> Tuple[Tensor, ...]:
        """
        Performs Kalman filtering given observations.

        Parameters
        ----------
        targets
            Observations, shape (batch_size, seq_length, output_dim)
        observed
            Flag tensor indicating which observations are genuine (1.0) and
            which are missing (0.0)

        Returns
        -------
        Tensor
            Log probabilities, shape (batch_size, seq_length)
        Tensor
            Mean of p(l_T | l_{T-1}), where T is seq_length, with shape
            (batch_size, latent_dim)
        Tensor
            Covariance of p(l_T | l_{T-1}), where T is seq_length, with shape
            (batch_size, latent_dim, latent_dim)
        """
        F = self.F
        # targets[t]: (batch_size, obs_dim)
        targets = _safe_split(
            targets, axis=1, num_outputs=self.seq_length, squeeze_axis=True
        )

        log_p_seq = []

        mean = self.prior_mean
        cov = self.prior_cov

        observed = (
            _safe_split(
                observed,
                axis=1,
                num_outputs=self.seq_length,
                squeeze_axis=True,
            )
            if observed is not None
            else None
        )

        for t in range(self.seq_length):
            # Compute the filtered distribution
            #   p(l_t | z_1, ..., z_{t + 1})
            # and log - probability
            #   log p(z_t | z_0, z_{t - 1})
            filtered_mean, filtered_cov, log_p = kalman_filter_step(
                F,
                target=targets[t],
                prior_mean=mean,
                prior_cov=cov,
                emission_coeff=self.emission_coeff[t],
                residual=self.residuals[t],
                noise_std=self.noise_std[t],
                latent_dim=self.latent_dim,
                output_dim=self.output_dim,
            )

            log_p_seq.append(log_p.expand_dims(axis=1))

            # Mean of p(l_{t+1} | l_t)
            mean = F.linalg_gemm2(
                self.transition_coeff[t],
                (
                    filtered_mean.expand_dims(axis=-1)
                    if observed is None
                    else F.where(
                        observed[t], x=filtered_mean, y=mean
                    ).expand_dims(axis=-1)
                ),
            ).squeeze(axis=-1)

            # Covariance of p(l_{t+1} | l_t)
            cov = F.linalg_gemm2(
                self.transition_coeff[t],
                F.linalg_gemm2(
                    (
                        filtered_cov
                        if observed is None
                        else F.where(observed[t], x=filtered_cov, y=cov)
                    ),
                    self.transition_coeff[t],
                    transpose_b=True,
                ),
            ) + F.linalg_gemm2(
                self.innovation_coeff[t],
                self.innovation_coeff[t],
                transpose_a=True,
            )

        # Return sequence of log likelihoods, as well as
        # final mean and covariance of p(l_T | l_{T-1} where T is seq_length
        return F.concat(*log_p_seq, dim=1), mean, cov

    def sample(
        self, num_samples: Optional[int] = None, scale: Optional[Tensor] = None
    ) -> Tensor:
        r"""
        Generates samples from the LDS: p(z_1, z_2, \ldots, z_{`seq_length`}).

        Parameters
        ----------
        num_samples
            Number of samples to generate
        scale
            Scale of each sequence in x, shape (batch_size, output_dim)

        Returns
        -------
        Tensor
            Samples, shape (num_samples, batch_size, seq_length, output_dim)
        """
        F = self.F

        # Note on shapes: here we work with tensors of the following shape
        # in each time step t: (num_samples, batch_size, dim, dim),
        # where dim can be obs_dim or latent_dim or a constant 1 to facilitate
        # generalized matrix multiplication (gemm2)

        # Sample observation noise for all time steps
        # noise_std: (batch_size, seq_length, obs_dim, 1)
        noise_std = F.stack(*self.noise_std, axis=1).expand_dims(axis=-1)

        # samples_eps_obs[t]: (num_samples, batch_size, obs_dim, 1)
        samples_eps_obs = _safe_split(
            Gaussian(noise_std.zeros_like(), noise_std).sample(num_samples),
            axis=-3,
            num_outputs=self.seq_length,
            squeeze_axis=True,
        )

        # Sample standard normal for all time steps
        # samples_eps_std_normal[t]: (num_samples, batch_size, obs_dim, 1)
        samples_std_normal = _safe_split(
            Gaussian(noise_std.zeros_like(), noise_std.ones_like()).sample(
                num_samples
            ),
            axis=-3,
            num_outputs=self.seq_length,
            squeeze_axis=True,
        )

        # Sample the prior state.
        # samples_lat_state: (num_samples, batch_size, latent_dim, 1)
        # The prior covariance is observed to be slightly negative definite
        # whenever there is excessive zero padding at the beginning of the time
        # series. We add positive tolerance to the diagonal to avoid numerical
        # issues. Note that `jitter_cholesky` adds positive tolerance only if
        # the decomposition without jitter fails.
        state = MultivariateGaussian(
            self.prior_mean,
            jitter_cholesky(
                F, self.prior_cov, self.latent_dim, float_type=np.float32
            ),
        )
        samples_lat_state = state.sample(num_samples).expand_dims(axis=-1)

        samples_seq = []
        for t in range(self.seq_length):
            # Expand all coefficients to include samples in axis 0
            # emission_coeff_t: (num_samples, batch_size, obs_dim, latent_dim)
            # transition_coeff_t:
            #   (num_samples, batch_size, latent_dim, latent_dim)
            # innovation_coeff_t: (num_samples, batch_size, 1, latent_dim)
            emission_coeff_t, transition_coeff_t, innovation_coeff_t = (
                _broadcast_param(coeff, axes=[0], sizes=[num_samples])
                if num_samples is not None
                else coeff
                for coeff in [
                    self.emission_coeff[t],
                    self.transition_coeff[t],
                    self.innovation_coeff[t],
                ]
            )

            # Expand residuals as well
            # residual_t: (num_samples, batch_size, obs_dim, 1)
            residual_t = (
                _broadcast_param(
                    self.residuals[t].expand_dims(axis=-1),
                    axes=[0],
                    sizes=[num_samples],
                )
                if num_samples is not None
                else self.residuals[t].expand_dims(axis=-1)
            )

            # (num_samples, batch_size, 1, obs_dim)
            samples_t = (
                F.linalg_gemm2(emission_coeff_t, samples_lat_state)
                + residual_t
                + samples_eps_obs[t]
            )
            samples_t = (
                samples_t.swapaxes(dim1=2, dim2=3)
                if num_samples is not None
                else samples_t.swapaxes(dim1=1, dim2=2)
            )
            samples_seq.append(samples_t)

            # sample next state: (num_samples, batch_size, latent_dim, 1)
            samples_lat_state = F.linalg_gemm2(
                transition_coeff_t, samples_lat_state
            ) + F.linalg_gemm2(
                innovation_coeff_t, samples_std_normal[t], transpose_a=True
            )

        # (num_samples, batch_size, seq_length, obs_dim)
        samples = F.concat(*samples_seq, dim=-2)
        return (
            samples
            if scale is None
            else F.broadcast_mul(
                samples,
                scale.expand_dims(axis=1).expand_dims(axis=0)
                if num_samples is not None
                else scale.expand_dims(axis=1),
            )
        )

    def sample_marginals(
        self, num_samples: Optional[int] = None, scale: Optional[Tensor] = None
    ) -> Tensor:
        r"""
        Generates samples from the marginals p(z_t),
        t = 1, \ldots, `seq_length`.

        Parameters
        ----------
        num_samples
            Number of samples to generate
        scale
            Scale of each sequence in x, shape (batch_size, output_dim)

        Returns
        -------
        Tensor
            Samples, shape (num_samples, batch_size, seq_length, output_dim)
        """
        F = self.F

        state_mean = self.prior_mean.expand_dims(axis=-1)
        state_cov = self.prior_cov

        output_mean_seq = []
        output_cov_seq = []

        for t in range(self.seq_length):
            # compute and store observation mean at time t
            output_mean = F.linalg_gemm2(
                self.emission_coeff[t], state_mean
            ) + self.residuals[t].expand_dims(axis=-1)

            output_mean_seq.append(output_mean)

            # compute and store observation cov at time t
            output_cov = F.linalg_gemm2(
                self.emission_coeff[t],
                F.linalg_gemm2(
                    state_cov, self.emission_coeff[t], transpose_b=True
                ),
            ) + make_nd_diag(
                F=F, x=self.noise_std[t] * self.noise_std[t], d=self.output_dim
            )

            output_cov_seq.append(output_cov.expand_dims(axis=1))

            state_mean = F.linalg_gemm2(self.transition_coeff[t], state_mean)

            state_cov = F.linalg_gemm2(
                self.transition_coeff[t],
                F.linalg_gemm2(
                    state_cov, self.transition_coeff[t], transpose_b=True
                ),
            ) + F.linalg_gemm2(
                self.innovation_coeff[t],
                self.innovation_coeff[t],
                transpose_a=True,
            )

        output_mean = F.concat(*output_mean_seq, dim=1)
        output_cov = F.concat(*output_cov_seq, dim=1)

        L = F.linalg_potrf(output_cov)

        output_distribution = MultivariateGaussian(output_mean, L)

        samples = output_distribution.sample(num_samples=num_samples)

        return (
            samples
            if scale is None
            else F.broadcast_mul(samples, scale.expand_dims(axis=1))
        )


class LDSArgsProj(mx.gluon.HybridBlock):
    def __init__(
        self,
        output_dim: int,
        noise_std_bounds: ParameterBounds,
        innovation_bounds: ParameterBounds,
    ) -> None:
        super().__init__()
        self.output_dim = output_dim
        self.dense_noise_std = mx.gluon.nn.Dense(
            units=1, flatten=False, activation="sigmoid"
        )
        self.dense_innovation = mx.gluon.nn.Dense(
            units=1, flatten=False, activation="sigmoid"
        )
        self.dense_residual = mx.gluon.nn.Dense(
            units=output_dim, flatten=False
        )

        self.innovation_bounds = innovation_bounds
        self.noise_std_bounds = noise_std_bounds

    def hybrid_forward(self, F, x: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        noise_std = (
            self.dense_noise_std(x)
            * (self.noise_std_bounds.upper - self.noise_std_bounds.lower)
            + self.noise_std_bounds.lower
        )

        innovation = (
            self.dense_innovation(x)
            * (self.innovation_bounds.upper - self.innovation_bounds.lower)
            + self.innovation_bounds.lower
        )

        residual = self.dense_residual(x)

        return noise_std, innovation, residual


def kalman_filter_step(
    F,
    target: Tensor,
    prior_mean: Tensor,
    prior_cov: Tensor,
    emission_coeff: Tensor,
    residual: Tensor,
    noise_std: Tensor,
    latent_dim: int,
    output_dim: int,
):
    """
    One step of the Kalman filter.

    This function computes the filtered state (mean and covariance) given the
    linear system coefficients the prior state (mean and variance),
    as well as observations.

    Parameters
    ----------
    F
    target
        Observations of the system output, shape (batch_size, output_dim)
    prior_mean
        Prior mean of the latent state, shape (batch_size, latent_dim)
    prior_cov
        Prior covariance of the latent state, shape
        (batch_size, latent_dim, latent_dim)
    emission_coeff
        Emission coefficient, shape (batch_size, output_dim, latent_dim)
    residual
        Residual component, shape (batch_size, output_dim)
    noise_std
        Standard deviation of the output noise, shape (batch_size, output_dim)
    latent_dim
        Dimension of the latent state vector
    Returns
    -------
    Tensor
        Filtered_mean, shape (batch_size, latent_dim)
    Tensor
        Filtered_covariance, shape (batch_size, latent_dim, latent_dim)
    Tensor
        Log probability, shape (batch_size, )
    """
    # output_mean: mean of the target (batch_size, obs_dim)
    output_mean = F.linalg_gemm2(
        emission_coeff, prior_mean.expand_dims(axis=-1)
    ).squeeze(axis=-1)

    # noise covariance
    noise_cov = make_nd_diag(F=F, x=noise_std * noise_std, d=output_dim)

    S_hh_x_A_tr = F.linalg_gemm2(prior_cov, emission_coeff, transpose_b=True)

    # covariance of the target
    output_cov = F.linalg_gemm2(emission_coeff, S_hh_x_A_tr) + noise_cov

    # compute the Cholesky decomposition output_cov = LL^T
    L_output_cov = F.linalg_potrf(output_cov)

    # Compute Kalman gain matrix K:
    # K = S_hh X with X = A^T output_cov^{-1}
    # We have X = A^T output_cov^{-1} => X output_cov = A^T => X LL^T = A^T
    # We can thus obtain X by solving two linear systems involving L
    kalman_gain = F.linalg_trsm(
        L_output_cov,
        F.linalg_trsm(
            L_output_cov, S_hh_x_A_tr, rightside=True, transpose=True
        ),
        rightside=True,
    )

    # compute the error
    target_minus_residual = target - residual
    delta = target_minus_residual - output_mean

    # filtered estimates
    filtered_mean = prior_mean.expand_dims(axis=-1) + F.linalg_gemm2(
        kalman_gain, delta.expand_dims(axis=-1)
    )
    filtered_mean = filtered_mean.squeeze(axis=-1)

    # Joseph's symmetrized update for covariance:
    ImKA = F.broadcast_sub(
        F.eye(latent_dim), F.linalg_gemm2(kalman_gain, emission_coeff)
    )

    filtered_cov = F.linalg_gemm2(
        ImKA, F.linalg_gemm2(prior_cov, ImKA, transpose_b=True)
    ) + F.linalg_gemm2(
        kalman_gain, F.linalg_gemm2(noise_cov, kalman_gain, transpose_b=True)
    )

    # likelihood term: (batch_size,)
    log_p = MultivariateGaussian(output_mean, L_output_cov).log_prob(
        target_minus_residual
    )

    return filtered_mean, filtered_cov, log_p



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest
from gluonts.dataset.repository import dataset_names, get_dataset
from gluonts.dataset.util import forecast_start


def check_train_test_split(dataset):
    prediction_length = dataset.metadata.prediction_length

    train_end = {}
    for entry in dataset.train:
        assert (
            entry["item_id"] not in train_end
        ), f"item {entry['item_id']} is duplicate"
        train_end[entry["item_id"]] = forecast_start(entry)

    test_end = {}
    for entry in dataset.test:
        test_end[entry["item_id"]] = forecast_start(entry)

    for k in test_end:
        if k not in train_end:
            continue
        expected_end = train_end[k] + prediction_length * train_end[k].freq
        assert (
            test_end[k] >= expected_end
        ), f"test entry for item {k} ends at {test_end[k]} < {expected_end}"


@pytest.mark.skip
@pytest.mark.timeout(300)
@pytest.mark.parametrize("name", dataset_names)
def test_data_leakage(name):
    try:
        dataset = get_dataset(name)
    except RuntimeError:
        print(f"WARN dataset '{name}' could not be obtained")

    check_train_test_split(dataset)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from functools import partial

import numpy as np
import pytest

from gluonts.mx import DeepAREstimator
from gluonts.mx.distribution import PiecewiseLinearOutput, StudentTOutput
from gluonts.mx.trainer import Trainer

from gluonts.testutil.dummy_datasets import make_dummy_datasets_with_features

common_estimator_hps = dict(
    freq="D",
    prediction_length=3,
    trainer=Trainer(epochs=1, num_batches_per_epoch=1),
)


@pytest.mark.parametrize(
    "estimator, datasets",
    [
        # No features
        (
            partial(DeepAREstimator, **common_estimator_hps),
            make_dummy_datasets_with_features(),
        ),
        # Single static categorical feature
        (
            partial(
                DeepAREstimator,
                **common_estimator_hps,
                use_feat_static_cat=True,
                cardinality=[5],
            ),
            make_dummy_datasets_with_features(cardinality=[5]),
        ),
        # Multiple static categorical features
        (
            partial(
                DeepAREstimator,
                **common_estimator_hps,
                use_feat_static_cat=True,
                cardinality=[3, 10, 42],
            ),
            make_dummy_datasets_with_features(cardinality=[3, 10, 42]),
        ),
        # Multiple static categorical features (ignored)
        (
            partial(DeepAREstimator, **common_estimator_hps),
            make_dummy_datasets_with_features(cardinality=[3, 10, 42]),
        ),
        # Single dynamic real feature
        (
            partial(
                DeepAREstimator,
                **common_estimator_hps,
                use_feat_dynamic_real=True,
            ),
            make_dummy_datasets_with_features(num_feat_dynamic_real=1),
        ),
        # Multiple dynamic real feature
        (
            partial(
                DeepAREstimator,
                **common_estimator_hps,
                use_feat_dynamic_real=True,
            ),
            make_dummy_datasets_with_features(num_feat_dynamic_real=3),
        ),
        # Multiple dynamic real feature (ignored)
        (
            partial(DeepAREstimator, **common_estimator_hps),
            make_dummy_datasets_with_features(num_feat_dynamic_real=3),
        ),
        # Both static categorical and dynamic real features
        (
            partial(
                DeepAREstimator,
                **common_estimator_hps,
                use_feat_dynamic_real=True,
                use_feat_static_cat=True,
                cardinality=[3, 10, 42],
            ),
            make_dummy_datasets_with_features(
                cardinality=[3, 10, 42], num_feat_dynamic_real=3
            ),
        ),
        # Both static categorical and dynamic real features (ignored)
        (
            partial(DeepAREstimator, **common_estimator_hps),
            make_dummy_datasets_with_features(
                cardinality=[3, 10, 42], num_feat_dynamic_real=3
            ),
        ),
    ],
)
@pytest.mark.parametrize(
    "distr_output", [StudentTOutput(), PiecewiseLinearOutput(num_pieces=5)]
)
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
@pytest.mark.parametrize("impute_missing_values", [False, True])
def test_deepar_smoke(
    distr_output, estimator, datasets, dtype, impute_missing_values
):
    estimator = estimator(
        distr_output=distr_output,
        dtype=dtype,
        impute_missing_values=impute_missing_values,
    )
    dataset_train, dataset_test = datasets
    predictor = estimator.train(dataset_train)
    forecasts = list(predictor.predict(dataset_test))
    assert all([forecast.samples.dtype == dtype for forecast in forecasts])
    assert len(forecasts) == len(dataset_test)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np
import pandas as pd
import pytest

from gluonts.dataset.common import ListDataset

from gluonts.dataset.field_names import FieldName
from gluonts.transform import AddAggregateLags

expected_lags_rolling = {
    "prediction_length_2": {
        "train": np.array(
            [
                [0, 0, 0, 1, 1, 1.5, 2, 2.5, 3, 3.5],
                [0, 0, 0, 0, 0, 1, 1, 1.5, 2, 2.5],
                [0, 0, 0, 0, 0, 0, 0, 1, 1, 1.5],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ]
        ),
        "test": np.array(
            [
                [0, 0, 0, 1, 1, 1.5, 2, 2.5, 3, 3.5, 4.5, 5.5],
                [0, 0, 0, 0, 0, 1, 1, 1.5, 2, 2.5, 3, 3.5],
                [0, 0, 0, 0, 0, 0, 0, 1, 1, 1.5, 2, 2.5],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1.5],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ]
        ),
    },
    "prediction_length_1": {
        "train": np.array(
            [
                [0, 0, 0, 1, 1, 1.5, 2, 2.5, 3, 3.5],
                [0, 0, 0, 0, 0, 1, 1, 1.5, 2, 2.5],
                [0, 0, 0, 0, 0, 0, 0, 1, 1, 1.5],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ]
        ),
        "test": np.array(
            [
                [0, 0, 0, 1, 1, 1.5, 2, 2.5, 3, 3.5, 4.5],
                [0, 0, 0, 0, 0, 1, 1, 1.5, 2, 2.5, 3],
                [0, 0, 0, 0, 0, 0, 0, 1, 1, 1.5, 2],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ]
        ),
    },
}


valid_lags_rolling = {
    "prediction_length_2": [1, 2, 3, 4, 6],
    "prediction_length_1": [1, 2, 3, 4, 6],
}


@pytest.mark.parametrize("pred_length", [2, 1])
def test_agg_lags(pred_length):
    # create dummy dataset
    target = np.array([1, 1, 1, 2, 2, 3, 3, 4, 5, 6])
    start = pd.Period("01-01-2019 01:00:00", freq="1H")
    freq = "1H"
    ds = ListDataset(
        [{FieldName.TARGET: target, FieldName.START: start}], freq=freq
    )

    # 2H aggregate lags
    lags_2H = [1, 2, 3, 4, 6]

    add_agg_lags = AddAggregateLags(
        target_field=FieldName.TARGET,
        output_field="lags_2H",
        pred_length=pred_length,
        base_freq=freq,
        agg_freq="2H",
        agg_lags=lags_2H,
    )

    assert add_agg_lags.ratio == 2

    train_entry = next(add_agg_lags(iter(ds), is_train=True))
    test_entry = next(add_agg_lags(iter(ds), is_train=False))

    assert (
        add_agg_lags.valid_lags
        == valid_lags_rolling[f"prediction_length_{pred_length}"]
    )

    assert np.allclose(
        train_entry["lags_2H"],
        expected_lags_rolling[f"prediction_length_{pred_length}"]["train"],
    )

    assert np.allclose(
        test_entry["lags_2H"],
        expected_lags_rolling[f"prediction_length_{pred_length}"]["test"],
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List

from mxnet.gluon import nn

from gluonts.core.component import validated
from gluonts.mx import Tensor


class MLP(nn.HybridBlock):
    """
    Defines an MLP block.

    Parameters
    ----------
    layer_sizes
        number of hidden units per layer.

    flatten
        toggle whether to flatten the output tensor.

    activation
        activation function of the MLP, default is relu.
    """

    @validated()
    def __init__(
        self, layer_sizes: List[int], flatten: bool, activation="relu"
    ) -> None:
        super().__init__()
        self.layer_sizes = layer_sizes
        with self.name_scope():
            self.layers = nn.HybridSequential()
            for layer, layer_dim in enumerate(layer_sizes):
                self.layers.add(
                    nn.Dense(layer_dim, flatten=flatten, activation=activation)
                )

    def hybrid_forward(self, F, x: Tensor) -> Tensor:
        """
        Parameters
        ----------
        F
            A module that can either refer to the Symbol API or the NDArray
            API in MXNet.
        x
            Input tensor

        Returns
        -------
        Tensor
            Output of the MLP given the input tensor.
        """
        return self.layers(x)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

import torch

from gluonts.torch.distributions import DiscreteDistribution


@pytest.mark.parametrize(
    "values, probs",
    [
        # Both elements in the batch have same distribution parameters.
        (
            torch.tensor([[0.0, 1.0, 0.5], [0.0, 0.5, 1.0]]),
            torch.tensor([[0.5, 0.5, 0.0], [0.5, 0.0, 0.5]]),
        )
    ],
)
@pytest.mark.parametrize(
    "obs, rps",
    [
        (
            torch.tensor([[0.0], [1.0]]),
            torch.tensor([0.0000, 0.2500]),
            # (0.5 * (0.0 - 0.0) + 0.0 + (1.0 - 1.0) * (1.0 - 0.5)) / 2
            # (0.5 * (1.0 - 0.0) + 0.0 + 1.0 * (1.0 - 1.0)) / 2
        ),
        (
            torch.tensor([[0.5], [0.75]]),
            torch.tensor([0.1250, 0.1875]),
            # (0.5 * (0.5 - 0.0) + 0.0 + (1.0 - 1.0) * (1.0 - 0.5)) / 2
            # (0.5 * (0.75 - 0.0) + 0.0 + (1.0 - 1.0) * (1.0 - 0.75)) / 2
        ),
        (
            torch.tensor([[-10.0], [10.0]]),
            torch.tensor([2.5, 7.0]),
            # (0.5 * (0.0 + 10.0) + 0.0 + (1.0 - 1.0) * (1.0 + 10.0)) / 2
            # (0.5 * (10.0 - 0.0) + 0.0 + (1.0) * (10.0 - 1.0)) / 2
        ),
    ],
)
def test_rps(values, probs, obs, rps):
    discrete_distr = DiscreteDistribution(values=values, probs=probs)
    rps_computed = discrete_distr.rps(obs)
    assert all(rps == rps_computed)


@pytest.mark.parametrize(
    "values, probs",
    [
        # Duplicate values occur (i) only in the middle (ii) at the extremes
        (
            torch.tensor(
                [
                    [-1.0, 0.0, 0.0, 0.0, 2.0, 2.0, 5.0],
                    [-1.0, -1.0, 0.0, 0.0, 2.0, 5.0, 5.0],
                ]
            ),
            torch.tensor(
                [
                    [0.1, 0.12, 0.03, 0.15, 0.05, 0.15, 0.4],
                    [0.15, 0.05, 0.13, 0.12, 0.05, 0.27, 0.23],
                ]
            ),
        )
    ],
)
@pytest.mark.parametrize(
    "probs_adjusted",
    [
        torch.tensor(
            [
                [0.1, 0.0, 0.0, 0.3, 0.0, 0.2, 0.4],
                [0.0, 0.2, 0.0, 0.25, 0.05, 0.0, 0.5],
            ]
        ),
    ],
)
def test_probs_duplicate_values(values, probs, probs_adjusted):
    probs_adjusted_computed = DiscreteDistribution.adjust_probs(values, probs)
    assert all(probs_adjusted.view(-1) == probs_adjusted_computed.view(-1))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

import numpy as np

import gluonts.zebras as zb

target = np.arange(10)
feat = np.arange(20).reshape(2, 10)
static = np.arange(3)
index = zb.periods("2020", "D", 10)

tf = zb.time_frame(
    {"target": target, "feat": feat},
    static={"static": static},
    index=index,
    metadata={"x": 42},
)


def test_time_series():
    ts = zb.time_series(target, index=index, name="target", metadata={"x": 42})
    assert len(ts) == len(target)
    assert (ts == target).all()
    assert ts.tdim

    assert len(ts) == 10
    assert len(ts[:4]) == 4
    assert len(ts[-4:]) == 4

    assert ts.metadata == {"x": 42}
    assert ts.name == "target"


def test_time_frame():
    assert tf.eq_to(tf)
    assert len(tf) == 10
    assert len(tf[:4]) == 4
    assert len(tf[-4:]) == 4
    assert tf.tdims

    assert tf["target"].name == "target"
    assert np.array_equal(tf["target"], target)
    assert np.array_equal(tf["feat"], feat)

    assert tf.metadata == {"x": 42}

    tf2 = tf.stack(["target", "feat"], "stacked")
    assert not tf.eq_to(tf2)
    assert tf2.columns["stacked"].shape == (3, 10)
    assert tf2.columns["stacked"].shape == (3, 10)
    assert "target" not in tf2.columns
    assert "feat" not in tf2.columns
    assert tf2.metadata == {"x": 42}
    assert tf2.tdims

    tf3 = tf.like({"foo": np.full(10, 7)})
    assert tf3.columns["foo"].shape == (10,)
    assert "target" not in tf3.columns
    assert "feat" not in tf3.columns
    assert tf3.metadata == {"x": 42}
    assert tf3.tdims


@pytest.mark.parametrize("split_index", [4, -6, index[4]])
def test_time_frame_split(split_index):
    sf = tf.split(split_index)
    assert len(sf) == 10
    assert len(sf.past) == 4
    assert len(sf.future) == 6
    assert sf.metadata == {"x": 42}
    assert sf.tdims

    sf = tf.split(split_index, past_length=2, future_length=2)
    assert len(sf) == 4
    assert len(sf.past) == 2
    assert len(sf.future) == 2
    assert np.array_equal(tf.static, sf.static)
    assert sf.metadata == {"x": 42}
    assert sf.tdims

    sf = tf.split(split_index, past_length=5, future_length=8)
    assert len(sf.past) == 5
    assert len(sf.future) == 8
    assert np.array_equal(tf.static, sf.static)
    assert sf.metadata == {"x": 42}
    assert sf.tdims


def test_time_frame_split_out_of_bounds():
    # Check extreme cases work, where either past length or future length is
    # empty.
    assert len(tf.split(0).past) == 0
    assert len(tf.split(len(tf)).future) == 0

    assert len(tf.split(tf.index.start).past) == 0
    assert len(tf.split(tf.index.start + len(tf)).future) == 0

    with pytest.raises(ValueError):
        tf.split(len(tf) + 1)

    with pytest.raises(Exception):
        tf.split(tf.index.start + len(tf) + 1)

    with pytest.raises(ValueError):
        tf.split(-(len(tf) + 1))

    with pytest.raises(Exception):
        tf.split(tf.index.start - 1)


def test_time_frame_resize():
    # lengthen
    tf2 = tf.resize(15, pad_value=99)
    assert len(tf2) == 15
    assert (tf2["target"][:5] == 99).all()

    tf2 = tf.resize(15, pad_value=99, pad="l")
    assert len(tf2) == 15
    assert (tf2["target"][:5] == 99).all()

    tf2 = tf.resize(15, pad_value=99, pad="r")
    assert len(tf2) == 15
    assert (tf2["target"][-5:] == 99).all()

    # shorten
    tf2 = tf.resize(5, pad_value=99)
    assert len(tf2) == 5
    assert (tf2["target"] == np.arange(5)).all()

    tf2 = tf.resize(5, pad_value=99, skip="r")
    assert len(tf2) == 5
    assert (tf2["target"] == np.arange(5)).all()

    tf2 = tf.resize(5, pad_value=99, skip="l")
    assert len(tf2) == 5
    assert (tf2["target"] == np.arange(5, 10)).all()


def test_split_frame_resize():
    sf = tf.split(5)

    sf2 = sf.resize(past_length=10, future_length=10, pad_value=99)
    assert len(sf2) == 20

    assert (sf2.past["target"][:5] == 99).all()
    assert (sf2.future["target"][-5:] == 99).all()

    sf2 = sf.resize(past_length=1, future_length=1, pad_value=99)

    assert sf2.past["target"][0] == 4
    assert sf2.future["target"][0] == 5


def test_split_frame_empty():
    sf = zb.split_frame()

    assert len(sf) == 0
    assert len(sf.past) == 0
    assert len(sf.future) == 0

    sf = zb.split_frame(past={"x": [1, 2, 3]})
    assert len(sf) == 3
    assert len(sf.past) == 3
    assert len(sf.future) == 0

    sf = zb.split_frame(future={"x": [1, 2, 3]})
    assert len(sf) == 3
    assert len(sf.past) == 0
    assert len(sf.future) == 3


def test_rename():
    tf2 = tf.rename({"a": "target"}, b="feat")

    assert tf2.columns.keys() == {"a", "b"}

    tf2 = tf.rename_static({"x": "static"})
    assert tf2.static.keys() == {"x"}

    tf2 = tf.rename_static(x="static")
    assert tf2.static.keys() == {"x"}



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import mxnet as mx
import numpy as np
from mxnet import nd

from gluonts.mx.block.cnn import CausalConv1D


def compute_causalconv1d(
    x: np.ndarray, kernels: np.ndarray, dilation: int
) -> np.ndarray:
    """
    Naive way to compute the 1-d causal convolution

    Parameters:
    x: np.array
      input array
    kernels: np.array
      array of weights, [w_1, ..., w_n] where n is the kernel_size
    dilation: int
      dilation rate d > 0, when d = 1, it reduces to the regular convolution
    Returns:
      Causal 1d convolution between x and kernels with the given dilation rate
    """

    conv_x = np.zeros_like(x)
    # compute in a naive way.
    for t, xt in enumerate(x):
        dial_offset = 0
        for i in reversed(range(len(kernels))):
            xt_lag = x[t - dial_offset] if t - dial_offset >= 0 else 0.0
            dial_offset += dilation
            conv_x[t] += kernels[i] * xt_lag

    return conv_x


def test_causal_conv_1d() -> None:
    """
    Here we test whether the causal conv1d matches the naive computation.
    """
    x = nd.random.normal(0, 1, shape=(1, 1, 10))

    conv1d = CausalConv1D(
        channels=1, kernel_size=3, dilation=3, activation=None
    )
    conv1d.collect_params().initialize(
        mx.init.One(), ctx=mx.cpu(), force_reinit=True
    )

    y1 = conv1d(x).reshape(shape=(-1,)).asnumpy()
    y2 = compute_causalconv1d(
        x=x.reshape(shape=(-1,)).asnumpy(),
        kernels=np.asarray([1.0] * 3),
        dilation=3,
    )

    assert (
        np.max(np.abs(y1 - y2)) < 1e-5
    ), "1d Causal Convolution calculation incorrect!"



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from dataclasses import dataclass
from functools import partial
from typing import Collection, Optional
from typing_extensions import Protocol, runtime_checkable

import numpy as np

from .evaluator import DirectEvaluator, DerivedEvaluator, Evaluator
from .aggregations import Mean, Sum
from .stats import (
    error,
    absolute_error,
    absolute_label,
    absolute_percentage_error,
    absolute_scaled_error,
    coverage,
    quantile_loss,
    scaled_interval_score,
    squared_error,
    symmetric_absolute_percentage_error,
)


@runtime_checkable
class Metric(Protocol):
    def __call__(self, axis: Optional[int] = None) -> Evaluator:
        raise NotImplementedError


def mean_absolute_label(axis: Optional[int] = None) -> DirectEvaluator:
    return DirectEvaluator(
        name="mean_absolute_label",
        stat=absolute_label,
        aggregate=Mean(axis=axis),
    )


def sum_absolute_label(axis: Optional[int] = None) -> DirectEvaluator:
    return DirectEvaluator(
        name="sum_absolute_label",
        stat=absolute_label,
        aggregate=Sum(axis=axis),
    )


@dataclass
class SumError:
    forecast_type: str = "0.5"

    def __call__(self, axis: Optional[int] = None) -> DirectEvaluator:
        return DirectEvaluator(
            name="sum_error",
            stat=partial(error, forecast_type=self.forecast_type),
            aggregate=Sum(axis=axis),
        )


@dataclass
class SumAbsoluteError:
    forecast_type: str = "0.5"

    def __call__(self, axis: Optional[int] = None) -> DirectEvaluator:
        return DirectEvaluator(
            name="sum_absolute_error",
            stat=partial(absolute_error, forecast_type=self.forecast_type),
            aggregate=Sum(axis=axis),
        )


@dataclass
class MSE:
    """Mean Squared Error"""

    forecast_type: str = "mean"

    def __call__(self, axis: Optional[int] = None) -> DirectEvaluator:
        return DirectEvaluator(
            name="MSE",
            stat=partial(squared_error, forecast_type=self.forecast_type),
            aggregate=Mean(axis=axis),
        )


@dataclass
class SumQuantileLoss:
    q: float

    def __call__(self, axis: Optional[int] = None) -> DirectEvaluator:
        return DirectEvaluator(
            name=f"sum_quantile_loss[{self.q}]",
            stat=partial(quantile_loss, q=self.q),
            aggregate=Sum(axis=axis),
        )


@dataclass
class Coverage:
    q: float

    def __call__(self, axis: Optional[int] = None) -> DirectEvaluator:
        return DirectEvaluator(
            name=f"coverage[{self.q}]",
            stat=partial(coverage, q=self.q),
            aggregate=Mean(axis=axis),
        )


@dataclass
class MAPE:
    """Mean Absolute Percentage Error"""

    forecast_type: str = "0.5"

    def __call__(self, axis: Optional[int] = None) -> DirectEvaluator:
        return DirectEvaluator(
            name="MAPE",
            stat=partial(
                absolute_percentage_error, forecast_type=self.forecast_type
            ),
            aggregate=Mean(axis=axis),
        )


@dataclass
class SMAPE:
    """Symmetric Mean Absolute Percentage Error"""

    forecast_type: str = "0.5"

    def __call__(self, axis: Optional[int] = None) -> DirectEvaluator:
        return DirectEvaluator(
            name="sMAPE",
            stat=partial(
                symmetric_absolute_percentage_error,
                forecast_type=self.forecast_type,
            ),
            aggregate=Mean(axis=axis),
        )


@dataclass
class MSIS:
    """Mean Scaled Interval Score"""

    alpha: float = 0.05

    def __call__(self, axis: Optional[int] = None) -> DirectEvaluator:
        return DirectEvaluator(
            name="MSIS",
            stat=partial(scaled_interval_score, alpha=self.alpha),
            aggregate=Mean(axis=axis),
        )


@dataclass
class MASE:
    """Mean Absolute Scaled Error"""

    forecast_type: str = "0.5"

    def __call__(self, axis: Optional[int] = None) -> DirectEvaluator:
        return DirectEvaluator(
            name="MASE",
            stat=partial(
                absolute_scaled_error, forecast_type=self.forecast_type
            ),
            aggregate=Mean(axis=axis),
        )


@dataclass
class ND:
    """Normalized Deviation"""

    forecast_type: str = "0.5"

    @staticmethod
    def normalized_deviation(
        sum_absolute_error: np.ndarray, sum_absolute_label: np.ndarray
    ) -> np.ndarray:
        return sum_absolute_error / sum_absolute_label

    def __call__(self, axis: Optional[int] = None) -> DerivedEvaluator:
        return DerivedEvaluator(
            name="ND",
            evaluators={
                "sum_absolute_error": SumAbsoluteError(
                    forecast_type=self.forecast_type
                )(axis=axis),
                "sum_absolute_label": sum_absolute_label(axis=axis),
            },
            post_process=self.normalized_deviation,
        )


@dataclass
class RMSE:
    """Root Mean Squared Error"""

    forecast_type: str = "mean"

    @staticmethod
    def root_mean_squared_error(mean_squared_error: np.ndarray) -> np.ndarray:
        return np.sqrt(mean_squared_error)

    def __call__(self, axis: Optional[int] = None) -> DerivedEvaluator:
        return DerivedEvaluator(
            name="RMSE",
            evaluators={
                "mean_squared_error": MSE(forecast_type=self.forecast_type)(
                    axis=axis
                )
            },
            post_process=self.root_mean_squared_error,
        )


@dataclass
class NRMSE:
    """RMSE, normalized by the mean absolute label"""

    forecast_type: str = "mean"

    @staticmethod
    def normalize_root_mean_squared_error(
        root_mean_squared_error: np.ndarray, mean_absolute_label: np.ndarray
    ) -> np.ndarray:
        return root_mean_squared_error / mean_absolute_label

    def __call__(self, axis: Optional[int] = None) -> DerivedEvaluator:
        return DerivedEvaluator(
            name="NRMSE",
            evaluators={
                "root_mean_squared_error": RMSE(
                    forecast_type=self.forecast_type
                )(axis=axis),
                "mean_absolute_label": mean_absolute_label(axis=axis),
            },
            post_process=self.normalize_root_mean_squared_error,
        )


@dataclass
class WeightedSumQuantileLoss:
    q: float

    @staticmethod
    def weight_sum_quantile_loss(
        sum_quantile_loss: np.ndarray, sum_absolute_label: np.ndarray
    ) -> np.ndarray:
        return sum_quantile_loss / sum_absolute_label

    def __call__(self, axis: Optional[int] = None) -> DerivedEvaluator:
        return DerivedEvaluator(
            name=f"weighted_sum_quantile_loss[{self.q}]",
            evaluators={
                "sum_quantile_loss": SumQuantileLoss(q=self.q)(axis=axis),
                "sum_absolute_label": sum_absolute_label(axis=axis),
            },
            post_process=self.weight_sum_quantile_loss,
        )


@dataclass
class MeanSumQuantileLoss:
    quantile_levels: Collection[float]

    @staticmethod
    def mean(**quantile_losses: np.ndarray) -> np.ndarray:
        stacked_quantile_losses = np.stack(
            [quantile_loss for quantile_loss in quantile_losses.values()],
            axis=0,
        )
        return np.ma.mean(stacked_quantile_losses, axis=0)

    def __call__(self, axis: Optional[int] = None) -> DerivedEvaluator:
        return DerivedEvaluator(
            name="mean_sum_quantile_loss",
            evaluators={
                f"quantile_loss[{q}]": SumQuantileLoss(q=q)(axis=axis)
                for q in self.quantile_levels
            },
            post_process=self.mean,
        )


@dataclass
class MeanWeightedSumQuantileLoss:
    quantile_levels: Collection[float]

    @staticmethod
    def mean(**quantile_losses: np.ndarray) -> np.ndarray:
        stacked_quantile_losses = np.stack(
            [quantile_loss for quantile_loss in quantile_losses.values()],
            axis=0,
        )
        return np.ma.mean(stacked_quantile_losses, axis=0)

    def __call__(self, axis: Optional[int] = None) -> DerivedEvaluator:
        return DerivedEvaluator(
            name="mean_weighted_sum_quantile_loss",
            evaluators={
                f"quantile_loss[{q}]": WeightedSumQuantileLoss(q=q)(axis=axis)
                for q in self.quantile_levels
            },
            post_process=self.mean,
        )


@dataclass
class MAECoverage:
    quantile_levels: Collection[float]

    @staticmethod
    def mean(
        quantile_levels: Collection[float], **coverages: np.ndarray
    ) -> np.ndarray:
        intermediate_result = np.stack(
            [np.abs(coverages[f"coverage[{q}]"] - q) for q in quantile_levels],
            axis=0,
        )
        return np.ma.mean(intermediate_result, axis=0)

    def __call__(self, axis: Optional[int] = None) -> DerivedEvaluator:
        return DerivedEvaluator(
            name="MAE_coverage",
            evaluators={
                f"coverage[{q}]": Coverage(q=q)(axis=axis)
                for q in self.quantile_levels
            },
            post_process=partial(
                self.mean, quantile_levels=self.quantile_levels
            ),
        )


@dataclass
class OWA:
    """Overall Weighted Average"""

    forecast_type: str = "0.5"

    @staticmethod
    def calculate_OWA(
        smape: np.ndarray,
        smape_naive2: np.ndarray,
        mase: np.ndarray,
        mase_naive2: np.ndarray,
    ) -> np.ndarray:
        return 0.5 * (smape / smape_naive2 + mase / mase_naive2)

    def __call__(self, axis: Optional[int] = None) -> DerivedEvaluator:
        return DerivedEvaluator(
            name="OWA",
            evaluators={
                "smape": SMAPE(forecast_type=self.forecast_type)(axis=axis),
                "smape_naive2": SMAPE(forecast_type="naive_2")(axis=axis),
                "mase": MASE(forecast_type=self.forecast_type)(axis=axis),
                "mase_naive2": MASE(forecast_type="naive_2")(axis=axis),
            },
            post_process=self.calculate_OWA,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

import numpy as np
import torch

from gluonts.torch.distributions.negative_binomial import NegativeBinomial


@pytest.mark.parametrize("probs", [0.05, 0.5, 0.8])
@pytest.mark.parametrize("total_count", [3, 7, 100])
def test_custom_neg_bin_logpdf_matches_scipy_neg_bin_logpdf(
    total_count, probs
):
    torch_dist = NegativeBinomial(total_count=total_count, probs=probs)
    scipy_dist = torch_dist.scipy_nbinom
    x = torch.randint(1, 20, (1000,))
    log_pdf_torch = torch_dist.log_prob(x).numpy()
    log_pdf_scipy = scipy_dist.logpmf(x.numpy())

    assert np.allclose(log_pdf_torch, log_pdf_scipy, rtol=1e-4)


@pytest.mark.parametrize("probs", [0.1, 0.5, 0.8])
@pytest.mark.parametrize("total_count", [3, 7, 100])
@pytest.mark.parametrize("value", [11.0, 5.0, 9.0])
def test_custom_neg_bin_cdf(total_count, probs, value):
    torch_dist = NegativeBinomial(total_count=total_count, probs=probs)
    scipy_dist = torch_dist.scipy_nbinom

    torch_cdf = torch_dist.cdf(torch.as_tensor(value)).numpy()
    scipy_cdf = scipy_dist.cdf(np.asarray(value))

    assert np.allclose(torch_cdf, scipy_cdf)


@pytest.mark.parametrize("probs", [0.1, 0.5, 0.8])
@pytest.mark.parametrize("total_count", [3, 7, 100])
@pytest.mark.parametrize("value", [0.1, 0.5, 0.9])
def test_custom_studentt_icdf(total_count, probs, value):
    torch_dist = NegativeBinomial(total_count=total_count, probs=probs)
    scipy_dist = torch_dist.scipy_nbinom

    torch_icdf = torch_dist.icdf(torch.as_tensor(value)).numpy()
    scipy_icdf = scipy_dist.ppf(np.asarray(value))

    assert np.allclose(torch_icdf, scipy_icdf)
